CROWD_PROP_NAMES = [
    "breakable",
    "fillable",
    "flammable",
    "openable", 
    "toggleable",
]

GPT_PROP_NAMES = [
    "cookable", 
    "heatSource",
    "coldSource",
    # "fireSource",       # deprecated
    "sliceable",
    "slicer",
    # "soakable",         # deprecated
    # "cleaningTool",     # deprecated
]

INTERNAL_PROP_NAMES = [
    "liquid",
    "visualSubstance",
    "microPhysicalSubstance",
    "macroPhysicalSubstance",
    "cloth",
    "rope",
    "softBody",
    "rigidBody",
    "assembleable",
    "meltable",
    "particleRemover",
    "particleApplier",
    "particleSource",
    "particleSink",
    "needsOrientation",
    "waterCook",
    "mixingTool",
    "sceneObject",
    "diceable",
]
